;===============================================================================
;  math functions
;===============================================================================

math_compare:
	;compare num1 to num2 and set Z (equal) and C (num1 < num2) flags
	ldi	ZL, num1
	ldi	count, DISPSIZE
math_compare_loop:
	ldd	temp2, Z+(num2-num1)
	ld	temp, Z+
	cp	temp, temp2
	brne	math_compare_end ;num1 != num2, flags already set by 'cp'
	dec	count
	brne	math_compare_loop
	;num1 == num2 (flags already set [Z=1 by last 'dec', C=0 by 'cp'])
math_compare_end:
	ret

;--------------------

math_packedinc:
	;increment packed BCD number in temp2
	inc	temp2
	subi	temp2, -6
	brhc	PC+2
	subi	temp2, 6
	ret

;--------------------

math_inc2:
	;increment num2 by 1 at place temp (0 = leftmost digit)
	mov	temp2, temp
	lsr	temp2
	ldi	ZL, num2
	add	ZL, temp2
math_inc2_loop:
	ld	temp2, Z
	sbrc	temp, 0
	swap	temp2
	subi	temp2, -0x10
	cpi	temp2, 0xA0
	brlo	math_inc2_end
	subi	temp2, 0xA0
	sbrc	temp, 0
	swap	temp2
	st	Z, temp2
	sbrs	temp, 0
	dec	ZL
	subi	temp, 1
	brcc	math_inc2_loop
	ori	flags, 1<<Error ;overflow
	ret
math_inc2_end:
	sbrc	temp, 0
	swap	temp2
	st	Z, temp2
	ret
	
;--------------------

math_dec2:
	;decrement num2 by 1 at place temp (0 = leftmost digit)
	;no overflow handling!
	mov	temp2, temp
	lsr	temp2
	ldi	ZL, num2
	add	ZL, temp2
math_dec2_loop:
	ld	temp2, Z
	sbrc	temp, 0
	swap	temp2
	subi	temp2, 0x10
	brcc	math_dec2_end
	subi	temp2, -0xA0
	sbrc	temp, 0
	swap	temp2
	st	Z, temp2
	sbrs	temp, 0
	dec	ZL
	subi	temp, 1
	brcc	math_dec2_loop
	;overflow
	ret
math_dec2_end:
	sbrc	temp, 0
	swap	temp2
	st	Z, temp2
	ret
	
;--------------------

math_shiftright_round2:
	;shift right (divide by 10) the number starting at ZL
	subi	ZL, -(DISPSIZE-1) ;pointer to last digit
	push	ZL
	;rounding
	ld	temp2, Z
	andi	temp2, 0x0F
	sbrc	flags, ShiftRounded
	subi	temp2, 1
	cpi	temp2, 5
	brlo	math_shiftright_shift ;round down
	ldi	temp, (DISPSIZE*2)-2 ;round up
	rcall	math_inc2
	ori	flags, 1<<ShiftRounded
	;shifting
math_shiftright_shift:
	pop	ZL
	ldi	count, DISPSIZE-1
math_shiftright_loop:
	ld	temp, Z
	swap	temp
	ld	temp2, -Z
	swap	temp2
	andi	temp, 0x0F
	andi	temp2, 0xF0
	or	temp, temp2
	std	Z+1, temp
	dec	count
	brne	math_shiftright_loop
	ld	temp2, Z ;clear first digit
	swap	temp2
	andi	temp2, 0x0F
	st	Z, temp2
	ret
	
;--------------------

math_shiftleft:
	;shift left (multiply by 10) the number starting at ZL
	ld	temp2, Z
	cpi	temp2, 0x10
	brsh	math_overflow
math_shiftleft_start:
	swap	temp2
	st	Z, temp2
	ldi	count, DISPSIZE-1
math_shiftleft_loop:
	ldd	temp2, Z+1
	swap	temp2
	std	Z+1, temp2
	andi	temp2, 0x0F
	ld	temp, Z
	andi	temp, 0xF0
	or	temp, temp2
	st	Z+, temp
	dec	count
	brne	math_shiftleft_loop
	ld	temp2, Z ;clear last digit
	andi	temp2, 0xF0
	st	Z, temp2
	ret
	
;--------------------

math_add:
	;addition: num1 = num1 + num2
	;handle negative numbers this way:
	;  same sign:       num1 = (num1's sign) (|num1| + |num2|)
	;  different signs: num1 = (num1's sign) (|num1| - |num2|)
	ldi	temp, 0
	sbrc	flags, num1neg
	inc	temp
	sbrc	flags, num2neg
	inc	temp
	cpi	temp, 1
	breq	math_sub_abs
	
math_add_abs:
	;add absolute values: num1 = (num1's sign) (|num1| + |num2|)
	clt ;T flag is used as decimal Carry Bit
	ldi	ZL, num1+DISPSIZE
	ldi	count, DISPSIZE
math_add_loop:
	ld	temp, -Z
	ldd	temp2, Z+(num2-num1)
	brtc	PC+2
	rcall	math_packedinc
	clt
	add	temp, temp2
	in	temp2, SREG
	push	temp2
	;check half-carry (lower nibble)
	brhs	math_add_halfcarry
	mov	temp2, temp
	andi	temp2, 0x0F
	cpi	temp2, 0x0A
	brlo	math_add_halfcarry_end
math_add_halfcarry:
	subi	temp, -0x06
	brcs	math_add_halfcarry_end
	pop	temp2
	rjmp	math_add_carry
math_add_halfcarry_end:
	pop	temp2
	;check carry (upper nibble)
	out	SREG, temp2
	brcs	math_add_carry
	cpi	temp, 0xA0
	brlo	math_add_carry_end
math_add_carry:
	subi	temp, -0x60
	cpi	count, 1
	breq	math_overflow
	set
math_add_carry_end:
	st	Z, temp
	dec	count
	brne	math_add_loop
	ret
	
;--------------------

math_overflow:
	ori	flags, 1<<Error
	ret
	
;--------------------
	
math_sub:
	;subtraction: num1 = num1 - num2, done by adding num1 and -num2
	ldi	temp, 1<<num2neg
	eor	flags, temp ;flip sign of num2
	rjmp	math_add

;--------------------

math_sub_abs:
	;subtract absolute values: num1 = (num1's sign) (|num1| - |num2|)
	rcall	math_compare
	brsh	math_sub_start
	;|num1| < |num2|: swap and flip sign of result
	ldi	ZL, num1
	ldi	count, DISPSIZE
math_sub_swap_loop:
	ld	temp, Z
	ldd	temp2, Z+(num2-num1)
	std	Z+(num2-num1), temp
	st	Z+, temp2
	dec	count
	brne	math_sub_swap_loop
	ldi	temp, 1<<num1neg
	eor	flags, temp
	
math_sub_start:
	;subtract |num2| from |num1| (at this point, |num1| >= |num2|)
	clt ;T flag is used as decimal Carry bit
	ldi	ZL, num1+DISPSIZE
	ldi	count, DISPSIZE
math_sub_loop:
	ld	temp, -Z
	ldd	temp2, Z+(num2-num1)
	brtc	PC+2
	rcall	math_packedinc
	clt
	sub	temp, temp2
	in	temp2, SREG
	;half-carry
	brhc	PC+2
	subi	temp, 0x06
	out	SREG, temp2
	;carry
	brcc	math_sub_carry_end
	subi	temp, 0x60
	set
math_sub_carry_end:
	st	Z, temp
	dec	count
	brne	math_sub_loop
	ret
	
;--------------------

math_mul:
	;multiply num1 and num2
	rcall	copy1to3 ;copy first factor to num3 and clear num1
	rcall	clearnum1
	
	ldi	temp, 1<<num1neg ;result's sign = (factor1's) XOR (factor2's)
	sbrc	flags, num3neg
	or	flags, temp
	sbrc	flags, num2neg
	eor	flags, temp
	
	;find first nonzero digit in second factor
	ldi	ZL, num2
	rcall	numfirst
	cpi	first, DISPSIZE*2
	brlo	PC+2
	ret ;second factor is 0 => result is 0
	push	first
	
	;shift second factor so second digit is first nonzero digit
	andi	flags, ~(1<<ShiftRounded)
	cpi	first, 0
	brne	math_mul_shiftleft
	ldi	ZL, num2
	rcall	math_shiftright_round2
	ldi	first, 1
	rjmp	math_mul_shiftend
math_mul_shiftleft:
	cpi	first, 1
	breq	math_mul_shiftend
	ldi	ZL, num2
	rcall	math_shiftleft
	dec	first
	rjmp	math_mul_shiftleft
math_mul_shiftend:
	
	;find first nonzero digit in first factor (num3)
	ldi	ZL, num3
	rcall	numfirst
	cpi	first, DISPSIZE*2
	brlo	math_mul_foundfirst
	;first factor is 0 => result is 0
	pop	temp ;remove data from stack
	ret
math_mul_foundfirst:
	push	first

	;add shifted 2nd factor to result (1st digit of 1st factor)-times,
	;then shift 2nd factor right 1 place and continue with next digit
	andi	flags, ~(1<<ShiftRounded)
	ldi	YL, num3
	mov	temp, first
	lsr	temp
	add	YL, temp
math_mul_add:
	ld	YH, Y
	sbrs	first, 0 ;odd digit -> right nibble
	swap	YH ;even digit -> left nibble
	andi	YH, 0x0F
	sbrc	first, 0 ;odd digit -> next digit is in the next byte
	inc	YL
	cpi	YH, 0
	breq	math_mul_add_next
math_mul_add_loop:
	rcall	math_add_abs
	dec	YH
	brne	math_mul_add_loop
math_mul_add_next:
	ldi	ZL, num2
	rcall	math_shiftright_round2
	inc	first
	cpi	first, DISPSIZE*2
	brlo	math_mul_add
	
	;shift result to correct position
	andi	flags, ~(1<<ShiftRounded)
	ldi	YH, DISPSIZE
	pop	temp
	sub	YH, temp
	pop	temp
	sub	YH, temp
	rjmp	math_muldiv_shiftresult
	
;--------------------

math_div:
	;divide num1 by num2
	rcall	clearnum3 ;clear num3 for result
	ldi	temp, 0 ;result's sign = (dividend's) XOR (divisor's)
	sbrc	flags, num1neg
	inc	temp
	sbrc	flags, num2neg
	inc	temp
	sbrc	temp, 0
	ori	flags, 1<<num3neg
	
	;find first nonzero digit in divisor
	ldi	ZL, num2
	rcall	numfirst
	cpi	first, DISPSIZE*2
	brlo	math_div_defined
	ori	flags, 1<<Error ;divisor is 0 => result is undefined
	ret
math_div_defined:
	push	first
	
	;shift divisor to the very left
math_div_shift1:
	cpi	first, 0
	breq	math_div_shift1end
	ldi	ZL, num2
	rcall	math_shiftleft
	dec	first
	brne	math_div_shift1
math_div_shift1end:
	
	;find first nonzero digit in dividend (num1)
	ldi	ZL, num1
	rcall	numfirst
	cpi	first, DISPSIZE*2
	brlo	math_div_foundfirst
	;dividend is 0 => result is 0
	pop	temp ;remove data from stack
	ret
math_div_foundfirst:
	push	first
	
	;shift dividend to the very left
math_div_shift2:
	cpi	first, 0
	breq	math_div_shift2end
	ldi	ZL, num1
	rcall	math_shiftleft
	dec	first
	brne	math_div_shift2
math_div_shift2end:

	;subtract while divident > divisor and increment result digit,
	;then shift divisor right and continue with next result digit
	andi	flags, ~(1<<ShiftRounded)
	ldi	YL, num3
	ldi	YH, 0x10
math_div_loop:
	rcall	math_compare
	brlo	math_div_next ;dividend < divisor
math_div_subtract: ;subtract divisor from dividend, increment result
	rcall	math_sub_abs
	ld	temp, Y
	add	temp, YH
	st	Y, temp
	rjmp	math_div_loop
math_div_next: ;continue with next digit
	ldi	ZL, num2
	rcall	math_shiftright_round2
	swap	YH ;toggle between adding 0x10 and 0x01
	sbrc	YH, 4 ;next byte if adding 0x10 comes next
	inc	YL
	cpi	YL, num3+DISPSIZE
	brlo	math_div_loop
	
	;copy result to num1 and shift result to correct position
	rcall	copy3to1
	andi	flags, ~(1<<ShiftRounded)
	ldi	YH, -(DISPSIZE-1)
	pop	temp
	sub	YH, temp
	pop	temp
	add	YH, temp
	rjmp	math_muldiv_shiftresult
	
;--------------------
	
math_muldiv_shiftresult:
	;shift result left YH times (YH < 0: shift right)
	brmi	math_muldiv_resultright
	brne	math_muldiv_resultleft
	ret
math_muldiv_resultright:
	ldi	ZL, num1
	rcall	math_shiftright_round2 ;round num2, so no effect on num1 here!
	inc	YH
	brne	math_muldiv_resultright
	ret
math_muldiv_resultleft:
	ldi	ZL, num1
	rcall	math_shiftleft
	dec	YH
	brne	math_muldiv_resultleft
	ret

;--------------------

math_inv:
	;calculate reciprocal inverse of num1
	rcall	copy1to2 ;copy num1 to num2
	rcall	clearnum1 ;load num1 with 1
	ldi	ZL, num1+(DISPSIZE/2-1)
	ldi	temp, 0x01
	st	Z, temp
	rcall	math_div ;calculate num1/num2 (1/x)
	ret

;--------------------

math_sqrt:
	;calculate square root of num1  (overflow if num1 >= [999...9])
	;  initialization: iter(0) = start = num1
	;  iteration: iter(i+1) = [start / iter(i) + iter(i)] / 2
	;error if num1 is negative
	sbrc	flags, num1neg
	rjmp	math_overflow
	;return 0 if num1 is 0
	rcall	clearnum2
	rcall	math_compare
	brne	PC+2
	ret
	
	;initialize values (start = num1, iter(0) = num1/2)
	rcall	copy1tostart
	ldi	ZL, num2+(DISPSIZE/2-1)
	ldi	temp, 0x02
	st	Z, temp
	rcall	math_div
	rcall	clearnum2 ;if num1/2 = 0, use iter(0) = start
	rcall	math_compare
	brne	PC+2
	rcall	copystartto1
	rcall	copy1toiter

	ldi	count, DISPSIZE*3 ;maximum number of iterations
math_sqrt_iteration:
	;iteration loop
	subi	count, 1
	brcs	math_sqrt_iteration_end
	push	count
	rcall	copystartto1
	rcall	copyiterto2
	rcall	math_div ;a / x(i)
	rcall	copyiterto2
	rcall	math_add ;... + x(i)
	rcall	clearnum2
	ldi	ZL, num2+(DISPSIZE/2-1)
	ldi	temp, 0x02
	st	Z, temp
	rcall	math_div ;... / 2
	rcall	copyiterto2
	rcall	copy1toiter
	rcall	math_compare ;iteration finished if x(i+1) = x(i)
	pop	count
	sbrs	flags, Error ;abort on error to prevent endless loop
	brne	math_sqrt_iteration
math_sqrt_iteration_end:
	ret
	